package net.sourceforge.chopchophttpclient;

import java.net.URI;
import java.util.Hashtable;

/**
 * @author rbeede
 *
 * <p style="font-weight: bold;">Notes:</p>
 * <ul>
 * 	<li>The content length for the collective may not be fully known until all ProgressListener members have
 *		registered themselves at least once by calling update.
 *	</li>
 * </ul>
 *
 */
public class CollectiveProgressListener implements ProgressListener {
	private final ProgressListener outPipeProgressListener;
	
	public CollectiveProgressListener(final ProgressListener outPipeProgressListener) {
		this.outPipeProgressListener = outPipeProgressListener;
	}
	
	// We track things by URI -> multiple ProgressListener (collective members) -> Long[2] with
	//	total read and piece content length
	public final Hashtable<URI,Hashtable<ProgressListener,Long[]>> uriToProgressListenerToLongTotals
		= new Hashtable<URI,Hashtable<ProgressListener,Long[]>>();
	
	/**
	 * Do not call this method.  Only individual progress listeners should call it.
	 * 
	 * @see net.sourceforge.chopchophttpclient.ProgressListener#update(java.net.URI, long, long)
	 * 
	 * @throws UnsupportedOperationException If this method is called.
	 */
	public void update(URI uri, long totalBytesRead, long contentLength) {
		throw new UnsupportedOperationException("This method should not be used.");
	}
	
	
	public ProgressListener getMember() {
		final CollectiveProgressListener me = this;
		
		return new ProgressListener() {
			public void update(URI uri, long totalBytesRead, long contentLength) {
				me.update(uri, this, totalBytesRead, contentLength);
			}
		};
	}
	

	protected void update(URI uri, ProgressListener inPipeProgressListener, long totalBytesRead, long contentLength) {
		if(null == this.outPipeProgressListener)  return;  // Don't waste time on nothing
		
		Hashtable<ProgressListener,Long[]> progressListenerToLongTotals;
		
		progressListenerToLongTotals = this.uriToProgressListenerToLongTotals.get(uri);
		
		if(null == progressListenerToLongTotals) {
			progressListenerToLongTotals = new Hashtable<ProgressListener,Long[]>();
			this.uriToProgressListenerToLongTotals.put(uri, progressListenerToLongTotals);
		}
		
		
		//Update the totals for this inPipeProgressListener
		progressListenerToLongTotals.put(inPipeProgressListener, new Long[] {totalBytesRead, contentLength});
		
	
		//Now tally up the current totals and send the update to our outPipeProgressListener
		long collectiveBytesRead = 0;
		long collectiveContentLength = 0;
		
		for(Long[] counts : progressListenerToLongTotals.values()) {
			collectiveBytesRead += counts[0];
			collectiveContentLength += counts[1];
		}
		
		//Send the collective total now
		this.outPipeProgressListener.update(uri, collectiveBytesRead, collectiveContentLength);
	}
}
