package net.sourceforge.chopchophttpclient.proxy;

import java.io.IOException;
import java.net.InetSocketAddress;
import java.net.Proxy;
import java.net.ProxySelector;
import java.net.SocketAddress;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.ArrayList;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

/**
 * Uses Java system properties or Linux proxy environment variables to determine an appropriate
 * proxy server setting for a given URI in the following preference order:
 * 
 * <ol>
 * <li>Environment variable http_proxy <i>order changes based on scheme with https</i></li>
 * <li>Environment variable https_proxy</li>
 * <li>Environment variable socks_proxy</li>
 * <li>System property http.proxyHost <i>order changes based on scheme with https</i></li>
 * <li>System property http.proxyPort</li>
 * <li>System property https.proxyHost</li>
 * <li>System property https.proxyPort</li>
 * <li>System property socks.proxyHost</li>
 * <li>System property socks.proxyPort</li>
 * </ol>
 * 
 * <p style="font-weight: bold;">Note:</p> The URI scheme determines whether http or https proxies are given first
 * 
 * <p style="font-weight: bold;">Note:</p> If the scheme is https but only http proxies are defined then the http
 * 		proxy will be provided instead.
 * 
 * <p style="font-weight: bold;">Note:</p>
 * The environment variables http_proxy, https_proxy, and socks_proxy are expected to follow the standard
 * Linux convention.  Also note that on Linux those environment variables are case sensitive and should be 
 * all <b>lowercase</b> instead of uppercase.  See man wget "Proxies" for details.
 * 
 * <p style="font-weight: bold;">Note:</p>
 * If no proxy port is specified then the default of {@link #DEFAULT_PROXY_PORT} is used.
 * 
 * <p style="font-weight: bold;">Note:</p>
 * no_proxy variable is not currently supported.
 * 
 * <p style="font-weight: bold;">Note:</p> 
 * Proxy username and password credentials are not returned.
 * 
 * <p style="font-weight: bold;">Example:</p>
 *
 * export "http_proxy=http://username:password@my-proxy.com:3128"<br />
 * export "https_proxy=https://username:password@my-secure-proxy.com:8080"<br />
 * export "socks_proxy=socks://username:password@my-socks-proxy.com:1080"<br />
 * 
 * 
 * @author rbeede
 *
 */
public class EnvironmentProxySelector extends ProxySelector {
	public static final int DEFAULT_PROXY_PORT = 3128;
	

	@Override
	public void connectFailed(URI uri, SocketAddress sa, IOException ioe) {
		//No reordering of proxies will be performed
	}

	@Override
	public List<Proxy> select(URI uri) {
		final String scheme = (null == uri) ? null : uri.getScheme().toLowerCase();
		
		final List<Proxy> proxyList = new ArrayList<Proxy>();
		
		
		//Start with Linux style environment variables
		
		//Order check (and thus proxies) by scheme if applicable
		final String[] envVariables;
		if("https".equals(scheme)) {
			envVariables = new String[] {
					"https_proxy",
					"http_proxy",
					"socks_proxy",
			};
		} else {
			envVariables = new String[] {
					"http_proxy",
					"https_proxy",
					"socks_proxy",
			};
		}

		for(String envVar : envVariables) {
			Proxy proxy = parseEnvVarProxy(envVar);
			
			if(null != proxy) {
				proxyList.add(proxy);
			}
		}
		
		
		final Map<String,String> systemPropHostToPortMap = new LinkedHashMap<String,String>();  //insertion order
		//Insert check order based on scheme to give priority to scheme
		if("https".equals(scheme)) {
			systemPropHostToPortMap.put("https.proxyHost", "https.proxyPort");
			systemPropHostToPortMap.put("http.proxyHost", "http.proxyPort");
			systemPropHostToPortMap.put("socks.proxyHost", "socks.proxyPort");
		} else {
			systemPropHostToPortMap.put("http.proxyHost", "http.proxyPort");
			systemPropHostToPortMap.put("https.proxyHost", "https.proxyPort");
			systemPropHostToPortMap.put("socks.proxyHost", "socks.proxyPort");
		}
		
		
		for(String hostPropName : systemPropHostToPortMap.keySet()) {
			String portPropName = systemPropHostToPortMap.get(hostPropName);
			

			String hostValue = System.getProperty(hostPropName);
			
			if(null == hostValue)  continue;
			
			int portValue;
			try {
				portValue = Integer.parseInt(System.getProperty(portPropName));
			} catch(NumberFormatException e) {
				portValue = DEFAULT_PROXY_PORT;
			} catch(NullPointerException e) {
				portValue = DEFAULT_PROXY_PORT;
			}
			
		
			SocketAddress address = new InetSocketAddress(hostValue, portValue);
			Proxy.Type proxyType;
			
			if(hostPropName.toLowerCase().startsWith(Proxy.Type.HTTP.toString().toLowerCase())) {
				proxyType = Proxy.Type.HTTP;
			} else if(hostPropName.toLowerCase().startsWith(Proxy.Type.SOCKS.toString().toLowerCase())) {
				proxyType = Proxy.Type.SOCKS;
			} else {
				//Invalid format
				continue;
			}
			
			
			proxyList.add(new Proxy(proxyType, address));
		}
		
		
		if(proxyList.isEmpty()) {
			//Required by ProxySelector.select method
			proxyList.add(Proxy.NO_PROXY);
		}
		
		return proxyList;
	}
	
	
	/**
	 * @param variableName The name of the environment variable
	 * @return The parsed environment variable value as a Proxy object.  Null if the value was invalid/non-existent.
	 */
	private Proxy parseEnvVarProxy(final String variableName) {
		final String envValue = System.getenv(variableName);
		
		if(null == envValue)  return null;
		

		final URI parsedURI;
		try {
			parsedURI = new URI(envValue);
		} catch (URISyntaxException e) {
			return null;
		}
		
		
		final SocketAddress address;
		if(-1 != parsedURI.getPort()) {
			address = new InetSocketAddress(parsedURI.getHost(), parsedURI.getPort());
		} else {
			address = new InetSocketAddress(parsedURI.getHost(), DEFAULT_PROXY_PORT);
		}
		
		Proxy.Type proxyType;
		
		if(parsedURI.getScheme().toLowerCase().startsWith(Proxy.Type.HTTP.toString().toLowerCase())) {
			proxyType = Proxy.Type.HTTP;
		} else if(parsedURI.getScheme().toLowerCase().startsWith(Proxy.Type.SOCKS.toString().toLowerCase())) {
			proxyType = Proxy.Type.SOCKS;
		} else {
			return null;
		}
		
		return new Proxy(proxyType, address);
	}


}
