package net.sourceforge.chopchophttpclient.util;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;

/**
 * This class writes the first memCapacity bytes to a byte array in memory and any further writes beyond that to
 * a temporary file on disk.  The temporary file will be marked for deletion with {@link File#deleteOnExit()} and
 * if the instantiated object has {@link #finalize()} called during garbage collection the {@link File#delete()}
 * will be called.
 * 
 * <p>
 * A temporary zero byte file is always created upon instantiation of this class as an object. 
 * </p>
 * 
 * <p>
 * Use toByteMultiArray() or writeTo() to get all the bytes.  Order will always be bytes in memory first followed
 * by bytes from the disk buffer.  If there are less bytes written than in memory than memCapacity then only the
 * number of actually written bytes will be returned.
 * </p>
 * 
 * <p>
 * The maximum number of total bytes supported is defined by {@link #MAXIMUM_SIZE}.
 * </p>
 * 
 * <p>
 * The maximum possible size of the in-memory buffer is {@link Integer#MAX_VALUE}. 
 * </p>
 *  
 * @author rbeede
 * 
 * @version 2009-05-11
 */
public class MemoryAndDiskOutputStream extends OutputStream {
	/**
	 * Used for reading from the temporary file on disk
	 */
	public static final int BLOCK_BUFFER_SIZE = 64 * 1024;
	
	/**
	 * 2^40 or 1 tebibyte
	 */
	public static final long MAXIMUM_SIZE = 1099511627776L;
	
	protected final byte[] memBuffer;
	protected int memBufferCount = 0;
	
	/**
	 * Total number of bytes written to either memory or the disk cache
	 */
	protected long count = 0;
	
	private final File tmpDiskBuffer;
	private final FileOutputStream tmpDiskBufferFOS;
	
	
	/**
	 * Calling this constructor also creates the zero byte disk buffer file on disk
	 * 
	 * @param memCapacity Initial capacity for the in-memory buffer.  Must not be negative.
	 * @throws IOException If a temporary file for the on-disk buffer file could not be created
	 * @throws IllegalArgumentException If memCapacity is negative
	 */
	public MemoryAndDiskOutputStream(final int memCapacity) throws IOException {
        if (memCapacity < 0) {
            throw new IllegalArgumentException("Negative initial memory capacity: " + memCapacity);
        }  // no need to check if > since it is an int

		this.memBuffer = new byte[memCapacity];
		
		this.tmpDiskBuffer = File.createTempFile(MemoryAndDiskOutputStream.class.getSimpleName(), null);
		this.tmpDiskBuffer.deleteOnExit();
		
		//On some OS this will lock the file for us too
		this.tmpDiskBufferFOS = new FileOutputStream(this.tmpDiskBuffer);
	}
	
	
	@Override
	public void write(int b) throws IOException {
		if(this.count >= MAXIMUM_SIZE) {
			throw new IOException("Maximum size of " + MAXIMUM_SIZE + " has been reached");
		}
		
		//Room in the in-memory buffer?
		if(this.count < this.memBufferCount) {
			this.memBuffer[this.memBufferCount] = (byte) b;
		
			this.memBufferCount++;
		} else {
			//Need to use the on-disk buffer
			this.tmpDiskBufferFOS.write(b);
		}

		
		this.count++;
	}
	
	@Override
	public void write(byte[] b, int off, int len) throws IOException {
		if(this.count >= MAXIMUM_SIZE) {
			throw new IOException("Maximum size of " + MAXIMUM_SIZE + " has been reached");
		}

		final int inMemBufferLength;
		if(len > (this.memBuffer.length - this.memBufferCount)) {
			// We can only copy so much into memory, the rest will go to disk
			inMemBufferLength = (this.memBuffer.length - this.memBufferCount);
		} else {
			inMemBufferLength = len;  // It all fits in memory
		}
		
		if(0 != inMemBufferLength) {
			//We had room in the in-memory buffer
			System.arraycopy(b, off, this.memBuffer, this.memBufferCount, inMemBufferLength);
			this.memBufferCount += inMemBufferLength;
		}
		
		if(inMemBufferLength != len) {
			//Not all bytes made it into the in-memory buffer, put them to disk
			final int diskBufferLength = len - inMemBufferLength;
			final int diskBufferOffset = off + inMemBufferLength;
			
			this.tmpDiskBufferFOS.write(b, diskBufferOffset, diskBufferLength);
		}
		
		this.count += len;
	}

	@Override
	protected void finalize() throws Throwable {
		super.finalize();
		
		if(null != this.tmpDiskBufferFOS) {
			try {
				this.tmpDiskBufferFOS.close();
			} catch(IOException e) {
				;  //ignore since throwing away object anyway
			}
			
			if(null != this.tmpDiskBuffer) {
				this.tmpDiskBuffer.delete();  //if needed
			}
		}
	}
	
	/**
	 * <p style="font-weight: bold;">Notes</p>
	 * <ul>
	 * 	<li>The first dimension will always have a size of 1.  This is for future expansion of larger support possibly
	 * 		up to {@link Long#MAX_VALUE} bytes.
	 * 	</li>
	 * 	<li>The first byte is at [0][0][0]</li>
	 * 	<li>If count were {@link #MAXIMUM_SIZE} then the last byte would be at index
	 * 		[0][(Math.ceil({@value #MAXIMUM_SIZE} / {@value Integer#MAX_VALUE})) - 1][({@value #MAXIMUM_SIZE} % {@value Integer#MAX_VALUE}) - 1]
	 * 	</li>
	 * </ul>
	 * 
	 * @return Three dimensional array with all the bytes from all buffers.  Returned array is a copy and not an
	 * 			internal reference
	 * @throws IOException If the on disk buffer temporary file could not be read
	 */
	public byte[][][] toByteMultiArray() throws IOException {
		final int firstDimSize = 1;  // fixed for now since MAXIMUM_SIZE is 2^40 
		
		final int secondDimSize = (int) Math.ceil(this.count / (double) Integer.MAX_VALUE);
		
		//third dimension size will be consistent except for the very last one
		final int lastThirdDimSize = (int) Math.ceil(this.count % Integer.MAX_VALUE);
		
		
		final byte[][][] allBytes = new byte[firstDimSize][secondDimSize][];
		
		
		for(int fst = 0; fst < firstDimSize; fst++) {
			for(int snd = 0; snd < secondDimSize; snd++) {
				final int offset = fst * snd;
				
				if(fst == (firstDimSize - 1) && (snd == secondDimSize - 1)) {
					// Last one which will have a special size
					allBytes[fst][snd] = getBytes(offset, lastThirdDimSize); 
				} else {
					allBytes[fst][snd] = getBytes(offset, Integer.MAX_VALUE);
				}
			}
		}
		
		return allBytes;
	}
	
	
	/**
	 * @param outputStream Where to write the bytes to.  Never closed.
	 * @throws IOException If any errors occur reading/writing data
	 */
	public void writeTo(final OutputStream outputStream) throws IOException {
		byte[] readBuffer;
		
		long offset = 0;
		
		while(offset < this.count) {
			// Reading in blocks helps to prevent us from using large amounts of memory
			// readBuffer.length may be less than BLOCK_BUFFER_SIZE as getBytes(...) won't go past this.count 
			readBuffer = getBytes(offset, BLOCK_BUFFER_SIZE);
			
			outputStream.write(readBuffer);
			
			offset += readBuffer.length;
		}
	}
	
	
	/**
	 * @param offset
	 * @param length Number of bytes to retrieve (if greater than remaining is set to number remaining from offset)
	 * @return Byte array with specified data from either the in-memory or disk buffer or both.  Length is always
	 * 			the actual amount of data read and may be less than the length passed in
	 * @throws IOException If any errors occur reading the data from disk
	 * @throws IllegalArgumentException If offset exceeds the size of the data 
	 */
	private byte[] getBytes(final long offset, int length) throws IOException {
		if(offset > this.count) {
			throw new IllegalArgumentException("Offset of " + offset + " is beyond size of " + this.count);
		} else if(length > this.count - offset) {
			length = (int) (this.count - offset);  // Read until actual end, not beyond
		}
		
		final byte[] combinedBytes = new byte[length];
		

		int bytesCopiedFromMemory = 0;

		
		// Read in anything from the in-memory buffer, if offset falls inside that
		if(offset < this.memBufferCount) {
			int lengthInMemory;
			if((offset + length) > this.memBufferCount) {
				lengthInMemory = (int) (this.memBufferCount - offset);  // Can only read rest of in-memory length
			} else {
				lengthInMemory = length;
			}

			System.arraycopy(this.memBuffer, (int)offset, combinedBytes, 0, lengthInMemory);
			
			bytesCopiedFromMemory = lengthInMemory;
		}
		
		
		if(offset > this.memBufferCount || (offset + length) > this.memBufferCount) {
			//Must read from disk
			
			// The offset we need from disk may be at the given offset or may need adjusted if read from memory too
			final long offsetOnDisk;
			if(offset > this.memBufferCount) {
				// Offset started past in-memory data (if any) so adjust to offset on disk
				offsetOnDisk = offset - this.memBufferCount;
			} else if((offset + length) > this.memBufferCount) {
				offsetOnDisk = 0;  // Read part in memory, just reading rest from beginning of disk
			} else {
				throw new RuntimeException("unexpected behavior with offset < this.memBufferCount");
			}
			
			final int lengthOnDisk = length - bytesCopiedFromMemory;

			
			final FileInputStream fis;  // New instance for thread safety versus using InputStream.reset()
			try {
				fis = new FileInputStream(this.tmpDiskBuffer);
			} catch (FileNotFoundException e) {
				throw new RuntimeException("Temporary file with buffer data is missing!", e);
			}
			
			fis.skip(offsetOnDisk);
			
			fis.read(combinedBytes, bytesCopiedFromMemory, lengthOnDisk);
			
			fis.close();
		}
		
		
		return combinedBytes;
	}
	
	
	public long size() {
		return this.count;
	}
}
